/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Jan 21, 2005
 *
 */
package com.sunlabs.archsim.gasp;

import java.util.Iterator;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * @author ib27688
 *
 */
public class ArbiterAction extends Action 
						implements ArbitrationParticipant {

	protected static final int MAX_PRIORITY= 0;
	protected static final int MIN_PRIORITY= 10;
	
	/** Input used to release the arbiter */
	protected InputTerminal doneInput;
	
	/** Is there an outstanding request */
	protected boolean requestOut;

	/** Did this action receive a grant from the arbiter? */
	protected boolean requestGranted;
	
	/** Arbiter */
	protected Arbiter arbiter;
		
	/** 
	 * Constructor for the arbiter action. Only subclasses can be
	 * instantiated.
	 * A subclass must define functions that are attached to the output
	 * terminals.
	 * @param name the name for the action
	 */
	protected ArbiterAction(String name) {
		super(name);
		notifyStateIsActiveCommand= new StateActiveArbiterCommand();
		notifyStateIsInactiveCommand= new StateInactiveArbiterCommand();		
		arbiter= new Arbiter(this, new RequestGrantedCommand());
		requestOut= false;
		requestGranted= false;
		doneInput= null;
	} // ArbiterAction constructor
	
	protected ArbiterAction(String name, CompositeEntity g) {
		this(name);
		setParent(g);
	} // ArbiterAction constructor
	
	/** Attach a peer arbiterAction - peer arbiter actions are
	 * mutually exclusive.
	 * 
	 * @param a arbiter action to be attached
	 */
	public boolean attachPeer(ArbiterAction a) {
		return arbiter.merge(a);
	} // attachPeer
	
	
	/** Designate one of the terminals to be a done input.
	 *  The consistency check verifies whether the done input
	 *  has been set - i.e. is not null.
	 */
	public boolean setDoneInput(InputTerminal in) {
		// make sure this action is attached to this terminal
		Action a= in.getAction();
		if (a != this) return false;
		doneInput= in;
		return true;
	} // setCommonInput
	


	/** When a state is active and the action fires, make sure that the
	 * lock is set.
	 */
	protected class StateActiveArbiterCommand extends Command {
		public void execute(Object param) {
			Terminal src= (Terminal)param;
			if (journalAction) {
				State state= src.getState();
				director.journal("Action " + getAlias()
						+ ", id= " + getID() + " notifed that state "
						+ state.getAlias() + ", id= " + state.getID() 
						+ " is active");						
			}		
			if (src instanceof InputTerminal) {
				numActiveInputs++;
				// if common input became active, release the arbiter
				if ( requestGranted && src == doneInput ) {
					requestGranted= false;
					arbiter.release(ArbiterAction.this);
				} // if commonInput				
				// are all of the "input" states active			
				if (canFire()) {
					// record that a request has been issued
					requestOut= true;
					// issue a request
					arbiter.request(ArbiterAction.this);
				} // if can fire
			} // if instance of input terminal
			
		} // execute
		
	} // StateActiveArbiterCommand
	
	
	protected class RequestGrantedCommand extends Command {

		public void execute(Object params) {
			requestGranted= true;
			requestOut= false;
			// iterate though outputs and schedule firing
			Iterator ti= terminals.iterator();
			while (ti.hasNext()) {
				Terminal t= (Terminal) (ti.next());
				if (t instanceof OutputTerminal) {
					computeAndFire((OutputTerminal)t);
				}
				else if (t instanceof InputTerminal
						&& t.getRole() == Terminal.SELF_RESETTING_INPUT) {
					// deactivate all states attached to input terminals
					t.setStateInactive();
					if (journalAction) {
						director.journal("Action " + getAlias()
								+ ", id= " + getID()
								+ ", fired on terminal " 
								+ t.getAlias() + ", id= " + t.getID() 
								+ " to set state inactive");					
					} // if journal_action
				} // else if
			} // while has next
		} // execute
		
	} // RequestGrantedCommand
	
	/** Receive an ackowledgment that a stat is inactive
	 *  The arbiter action is monitoring it's commonState = when it becomes
	 */
	protected class StateInactiveArbiterCommand extends Command {
		public void execute(Object s) {
			Terminal src= (Terminal)s;
			if (journalAction) {
				State state= src.getState();
				director.journal("Action " + getAlias()
						+ ", id= " + getID() + " notifed that state "
						+ state.getAlias() + ", id= " + state.getID() 
						+ " is INactive");						
			}			
			if (src instanceof InputTerminal) {
				numActiveInputs--;
				if (requestOut) {
					// need to withdraw a request
					requestOut= false;
					arbiter.withdraw(ArbiterAction.this);
				}
			}// if InputTerminal
		}
	} // StateInactiveArbiterCommand

	
	/** Chek whether this action can fire. If so, pceed and fire.
	 */
	protected void fireIfPossible() {
		// use noifyStateIsActiveCommand - it will check whether all
		// of the inputs are valid and fire if that is the case
			if (canFire())
				notifyStateIsActiveCommand.execute(null);
	} // fireIfPossible
	
	
	protected class SetArbiterCommand extends Command {
		public void execute(Object params) {
			arbiter= (Arbiter)params;
		}
	}
	
	public Arbiter getArbiter() {
		return arbiter;
	}

	public void setArbiter(Arbiter a) {
		arbiter= a;
	}
	
	public void setArbiterDelay(Delay d) {
		arbiter.setDelay(d);
	} // setArbterDelay
	
	
	/** add additional checks specific to the ArbiterAction */
	public boolean selfCheck() {
		// first run generic checks
		boolean pass= super.selfCheck();
		
		pass= pass && doneInput != null;
		if (!pass) {
			director.error("Arbiter action " + getAlias()
					+ ",id= " + getID() + ", has commonInput undefined");
		}
		
		return pass;
	}
	
} // ArbiterAction

